# -*- coding: utf-8 -*-

# Standard Library
import json
from collections import defaultdict
# Vendor Library
from yadic.container import Injectable
from django.db.models import Q

from m3.actions import ApplicationLogicException

# Project Library
from kinder.webservice.helpers import (find_existing_child,
    find_existing_delegate)
from kinder.webservice.spyne_ws.exceptions import SpyneException
from kinder.core.helpers import prepare_name
from kinder.core.declaration.models import Declaration, DeclarationPrivilege
from kinder.core.declaration_status.models import (
    DeclarationStatus, DSS)
from kinder.core.unit.models import Unit
from kinder.core.queue_module.api import get_queue_by_context
from kinder.core.queue_module.context import QueueContext

from kinder.core.audit_log.models import AdditionalEncoder
from kinder.core.audit_log.helpers import get_model
from kinder.core.children.models import ChildrenDelegate
from concentrator.change import (
    DeclarationChangeHelper, DeclarationDocsChangeHelper, ChangeHelper,
    StorageHelper
)
from concentrator.exceptions import DuplicateExternalID
from concentrator.models import ChangeDeclaration
from concentrator.rules import *
from concentrator.webservice.spyne_objects_proxy import *
from concentrator.webservice.django_objects_proxy import *
from concentrator.webservice.helpers import (
    get_delegate, DocumentWorker, _get_document_reference_from_ctx)
from concentrator import settings
from kinder.provider import features
from .config import cont
from ..models import ChangeSource


class NewApplicationProxy(object):
    """ Прокси объект для метода NewApplication """

    def __init__(self, request, binary_data, req_code):

        self._request = request
        self._external_id = request.ExternalId
        self._submit_date = request.SubmitDate
        self._entry_date = request.EntryDate
        self._adaptation_program_type = request.AdaptationProgramType
        self._adaptation_program_doc_info = getattr(
            request, "AdaptationProgramDocInfo", None)
        self._delegate_data = request.Applicant
        self._child_data = request.DeclaredPerson
        self._benefits = self._child_data.Benefits
        self._edu_organizations = getattr(request, 'EduOrganizations', None)
        self._schedule = getattr(request, "Schedule", None)
        self._binary_data = binary_data
        self._request_code = req_code
        self._document_references = _get_document_reference_from_ctx(
            request)

        self._ch_helpers = []

    def _build_delegate(self):
        u"""Создает обьект представителя из данных запроса."""
        return DelegateProxy(self._delegate_data).get()

    def _build_child(self):
        u"""Создает обьект ребенка из данных запроса."""
        return ChildrenProxy(
            self._child_data,
            self._adaptation_program_type,
            self._adaptation_program_doc_info).get()

    def _create_benefits(self):
        benefits = getattr(self._benefits, "Benefit", None) or []
        privileges = PrivilegeProxy(
            [(benefit.name, benefit.Type)
                for benefit in benefits]).get()
        return privileges

    def _get_schedule_type(self):
        """Возвращает приоритетное время прибывания, либо None"""

        # В запросе может быть указано несколько режимов,
        # но в системе в заявке можно указать только одно.
        # Приоритет определяется следующим образом:
        # [полн., сокр., крат., продл., кругл.]
        priority = [3, 2, 1, 4, 5]
        schedule = getattr(
            self._schedule, 'ScheduleType', []
        ) if self._schedule else []
        schedule = sorted(schedule, key=lambda x: priority.index(x))

        return WorkTypeRule.get_system(schedule[0]) if schedule else None

    def _build_declaration(self):
        u"""Создает обьект заявления из данных запроса."""
        return Declaration(
            date=self._submit_date,
            desired_date=self._entry_date,
            offer_other=self._edu_organizations.AllowOfferOther,
            comment_privileges=getattr(self._benefits,
                                       'BenefitsDocInfo', ''),
            client_id=self._external_id,
            work_type_id=self._get_schedule_type(),
            source=features.get_feature(
                'concentrator.webservice.service.'
                'NewApplicationRequest.get_source')(),
            portal=1)

    def _attache_documents(self, declaration):

        DocumentWorker(
            declaration, self._binary_data, self._request_code).attach()

    def _save_doc_extra_info(self, declaration):

        DocExtraInfoProxy(declaration, self._document_references).save()

    def _save_privileges(self, privileges, declaration):
        for privilege in privileges:
            DeclarationPrivilege.objects.create(
                declaration=declaration,
                datetime=datetime.datetime.now(),
                privilege=privilege)

    def check_model(self, name_model, conc_data, have_data):
        """собирает все изменения модели
        @type name_model: str
        @param name_model: имя модели из MAP_CHANGES
        @param conc_data: spyne object инстанс модеи воссозданный
         из данных на изменение
        @param have_data: django model инстанс модели из хранилища
        """

        class_model = get_model(name_model)
        ch_help = ChangeHelper(class_model, name_model)
        ch_help.check_diff(conc_data, have_data)
        self._ch_helpers.append(ch_help)

    def _create_or_update_child(self, child):
        u"""Ищет ребенка среди существующих, если такой имеется,
        то добавляются только изменения для принятия"""
        existing_child = find_existing_child(child)

        if existing_child:
            self.check_model('Children', child, existing_child)
            child = existing_child

        return child

    def _create_or_update_delegate(self, delegate, delegate_extension, child):
        u"""Ищет представителя среди существующих, если такой имеется,
        то добавляются только изменения для принятия"""
        existing_delegate = find_existing_delegate(delegate, child)

        if existing_delegate:
            self.check_model('Delegate', delegate, existing_delegate)
            delegate = existing_delegate
            if (
                hasattr(delegate, 'concentrator_delegate')
                and delegate.concentrator_delegate
            ):
                delegate_extension = delegate.concentrator_delegate

        return delegate, delegate_extension

    def process(self):
        check_organization_priorities(self._edu_organizations)

        try:
            # Пытаемся найти в системе заявку с таким же ExternalID
            existing_declaration = Declaration.objects.get(
                client_id=self._external_id)
        except Declaration.DoesNotExist:
            # Если заявки с таким статусом нет, то начинаем процесс создания
            # заявки
            pass
        else:
            # Заявка с таким же ExternalID уже существует
            # Если её статус не "Отказано в услуге" или "Архивная"
            # То возвращаем ответ со статусом REJECT
            if existing_declaration.status.code not in [
                    DSS.REFUSED, DSS.ARCHIVE]:
                raise DuplicateExternalID(
                    u"Заявление с таким ExternalID уже существует в системе")

            # Сохраняем то, что пришло во вкладку "Изменения с ЕПГУ"
            update_proxy = cont.get("UpdateProxy", "for_new_application")
            update_proxy.process(
                self._request, existing_declaration,
                binary_data=self._binary_data,
                request_code=self._request_code)

            return existing_declaration.id

        declaration = self._build_declaration()

        child = self._create_or_update_child(self._build_child())

        delegate, delegate_extension = self._build_delegate()
        if child.id:  # ребенок был найден, пробуем обновить представителя
            delegate, delegate_extension = self._create_or_update_delegate(
                delegate, delegate_extension, child
            )

        privileges = self._create_benefits()
        mo = get_mo_from_request(self._edu_organizations, declaration)

        try:
            delegate.save()
        except ApplicationLogicException as err:
            raise SpyneException(message=err.message)

        delegate_extension.delegate = delegate
        delegate_extension.save()

        try:
            child.save()
        except ApplicationLogicException as err:
            raise SpyneException(message=err.message)

        ChildrenDelegate.objects.create(children=child, delegate=delegate)

        declaration.children = child
        declaration.mo = mo
        declaration.status = DeclarationStatus.objects.get(
            code=DSS.DUL_CONFIRMATING)
        declaration.save()

        # создаются изменения по заявке для принятия
        StorageHelper.create_change(declaration, self._ch_helpers)

        # в declaration в datetime полях на данный момент может хранится
        # offset-aware datetime, но при получении через get может вернуться
        # offset-naive datetime (в зависимости от настроек приложения) и
        # наоборот, это может привести к ошибкам при сравнении поэтому
        # обновляем переменную
        declaration = Declaration.objects.get(pk=declaration.id)

        declaration_units = get_declaration_units_from_request(
            self._edu_organizations, declaration)

        for declaration_unit in declaration_units:
            # bulk create не использую т.к. он не вызывает save
            # метода и не срабатывают сигналы post_save и pre_save
            declaration_unit.declaration = declaration
            declaration_unit.save()

        self._save_privileges(privileges, declaration)
        self._save_doc_extra_info(declaration)
        self._attache_documents(declaration)

        return declaration.id


class GetApplicationQueueProxy(object):
    """
    Прокси объект для метода GetApplicationQueue
        Запрос текущей очереди заявления
    Возвращает список очередей во все желаемые доу заявки,
    каждая очередь содержит все заявки до нашей.
    """

    def __init__(self, declaration, request):
        assert declaration
        assert isinstance(declaration, Declaration)
        self.declaration = declaration
        self.request = request

    @staticmethod
    def _create_application(declaration, i):
        return ApplicationProxy.create_application(declaration, i)

    @staticmethod
    def _create_queue(unit, count, app):
        conf = {
            'EduOrganizationCode': str(unit.id),
            'ApplicationsCount': count,
            'Application': app
        }
        q = Queue(**conf)
        return q

    def process(self):
        """
        1. Если AllApplications == false, значит это первый запрос и мы должны
        отдать информацию по текущей организации, и вернуть им признак
        разрешено у нас отдавать очередь или нет в тэге SupportAllApplications
        2. Если передано AllApplications == true, но у нас в настроках
        SupportAllApplications = false,
         то отрабатываем как при AllApplications == false
        3. Если передано AllApplications == true и у нас в настроках
        SupportAllApplications = true То отдаем всю очередь перед заявкой
        4. Если передан id организации, то фильтруем желаемые организации
         по нему
        """
        result = []
        decl_id = self.declaration.id
        all_app = self.request.AllApplications

        if settings.GET_APPQ_FULL:
            if all_app or all_app is None:
                all_app = True
            else:
                all_app = False
        else:
            all_app = False

        if self.declaration.status.code not in DSS.status_queue():
            return GetApplicationQueueResponse(
                Queue=result, SupportAllApplications=settings.GET_APPQ_FULL)

        qs = self.declaration.declarationunit_set.all()
        if self.request.EduOrganizationCode:
            qs = qs.filter(unit_id=self.request.EduOrganizationCode)
        qs = qs.order_by('ord')

        for d_unit in qs:
            unit = d_unit.unit
            # считаем очередь до заявки в желаемое доу
            ctx = QueueContext.make_base_context(
                self.declaration.children.date_of_birth,
                unit)
            app = []
            i = 0
            for i, decl in enumerate(get_queue_by_context(ctx), 1):
                if not all_app and decl_id != decl.id:
                    continue
                app.append(
                    GetApplicationQueueProxy.
                    _create_application(decl, i))
                if decl.id == decl_id:
                    break
            result.append(
                GetApplicationQueueProxy._create_queue(
                    unit, i, app))

        return GetApplicationQueueResponse(
            Queue=result, SupportAllApplications=settings.GET_APPQ_FULL)


class GetApplicationStateProxy(object):
    """
    Прокси объект для метода GetApplicationQueue
        Запрос текущего статуса заявления
    """

    def __init__(self, declaration):
        assert declaration
        assert isinstance(declaration, Declaration)
        self.declaration = declaration

    def process(self):
        status_code = DeclarationStatusCodeRule.get_concetr(
            self.declaration.status.code)
        return GetApplicationStateResponse(**{
            'Code': str(status_code) if status_code else None,
            'Name': DeclarationStatusNameRule.get_concetr(
                self.declaration.status.code),
            'Details': self.declaration.status.description
        })


class FindApplicationsProxy(object):

    BY_DELEGATE = 1
    BY_CHILD = 2

    DOC_TYPE_RULE = {
        BY_DELEGATE: DelegateDocTypeRule,
        BY_CHILD: ChildrenDocTypeRule
    }

    def __init__(self, person_info, mode=BY_DELEGATE):

        self._person = person_info
        self._search_dict = {}
        self._mode = mode

    def _fill_attribute(
            self, spyne_attribute, django_field):

        value = getattr(self._person, spyne_attribute, None)
        if value:
            self._search_dict[django_field] = value

    def _convert_doc_type(self):

        value = getattr(self._person, 'DocType', None)
        if value:
            self._person.DocType = self.DOC_TYPE_RULE[
                self._mode].get_system(
                    self._person.DocType)

    def _convert_data(self):

        self._convert_doc_type()

    def _find_declarations(self, prefix):

        self._convert_data()

        self._fill_attribute('FirstName', '%s__firstname' % prefix)
        self._fill_attribute('LastName', '%s__surname' % prefix)
        self._fill_attribute('MiddleName', '%s__patronymic' % prefix)
        self._fill_attribute('Snils', '%s__snils' % prefix)
        self._fill_attribute('DateOfBirth', '%s__date_of_birth' % prefix)
        self._fill_attribute('DocType', '%s__dul_type' % prefix)
        self._fill_attribute('DocNumber', '%s__dul_number' % prefix)
        self._fill_attribute('DocIssueDate', '%s__dul_date' % prefix)

        declarations = Declaration.objects.filter(
            **self._search_dict).distinct('id')
        return declarations

    def process(self):
        if self._mode == self.BY_CHILD:
            type_response = FindApplicationsByDeclaredPersonResponse
            declarations = self._find_declarations('children')
        else:
            type_response = FindApplicationsByApplicantResponse
            declarations = self._find_declarations(
                'children__childrendelegate__delegate')

        applicants = []
        for declaration in declarations:

            applicants.append(
                ApplicationSearchResultProxy.create_application(
                    declaration))

        return type_response(Application=applicants)


class GetApplicationProxy(object):
    """
    Прокси объект для метода GetApplication
        Получение данных Заявления для изменения
    """

    def __init__(self, declaration, decl_proxy, app_proxy, dou_proxy):
        assert declaration
        assert isinstance(declaration, Declaration)
        self.declaration = declaration
        self.decl_proxy = decl_proxy
        self.app_proxy = app_proxy
        self.dou_proxy = dou_proxy
        self.schedule_proxy = ScheduleDataProxy(declaration)

    def process(self):
        hn_confirmation = self.declaration.children.health_need_confirmation

        conf = {
            'SubmitDate': self.declaration.date,
            'EntryDate': get_desire_date(self.declaration),
            'EducationProgramType': '',
            'AdaptationProgramType': get_health_need(self.declaration),
            'AdaptationProgramDocInfo': hn_confirmation,
            'Applicant': self.app_proxy.create_spyne_obj(),
            'DeclaredPerson': self.decl_proxy.create_spyne_obj(),
            'Schedule': self.schedule_proxy.create_spyne_obj(),
            'EduOrganizationsData': self.dou_proxy.create_spyne_obj(),
            'ApplicationRules': ApplicationRulesDataProxy.create_rules_data(
                self.declaration),
            'DocumentReferences': DocumentReferencesDataProxy.get_data(
                self.declaration)
        }

        return GetApplicationResponse(**conf)


class UpdateProxy(NewApplicationProxy):
    """
    Прокси объект отвечает за сохранение запроса
    об изменении заявки
    """

    __metaclass__ = Injectable

    def __init__(
            self, request, declaration, binary_data, request_code, **kwargs):
        super(UpdateProxy, self).__init__(request, binary_data, request_code)
        if hasattr(request, "EduOrganizationsData"):
            self._edu_organizations = request.EduOrganizationsData
        self.declaration = declaration
        self.result = []

    def chek_model(self, name_model, conc_data, have_data):
        """собирает в список все ссылки на наследников класса
        @param name_model: str имя модели из MAP_CHANGES
        @param conc_data: spyne object инстанс модеи воссозданный
         из данных на изменение
        @param have_data: django model инстанс моделииз хранилища
        """
        class_model = get_model(name_model)
        class_helper = self.storage.get_change_helper(name_model)
        ch_help = class_helper(class_model, name_model)
        ch_help.check_diff(conc_data, have_data)
        self.result.append(ch_help)

    def _get_declaration_units_list(self):
        """Возвращает список (unit_id, ord) по данным из запроса,
        прошедших всю валидацию"""
        decl_units = get_declaration_units_from_request(
            self._edu_organizations, self.declaration)
        return map(lambda x: (x.unit_id, x.ord), decl_units)

    def _create_benefits(self):
        result = super(UpdateProxy, self)._create_benefits()
        return map(lambda x: x.id, result)

    def _check_benefits(self):
        conc_benefits = self._create_benefits()
        benefits = DeclarationPrivilege.objects.filter(
            declaration=self.declaration).values_list('privilege_id', flat=True)
        self.chek_model('DeclarationPrivilege', conc_benefits, benefits)

    def _check_docs(self):
        """
        - сохраняем бинарники к нам в базу с признаком не принято
        - помещаем данные о новых файлах в параметр класса
        DeclarationDocsChangeHelper
        :return:
        """
        if self._document_references:
            helper = DeclarationDocsChangeHelper(DocExtraInfo)
            helper.check_diff(
                self._document_references,
                self.declaration,
                self._binary_data,
                self._request_code)
            self.result.append(helper)

    def process(self):
        check_organization_priorities(self._edu_organizations)

        msg = u'OK'
        # проверем поля заявки
        self.chek_model('Declaration', self._build_declaration(),
                        self.declaration)
        # проверяем поля ребенка
        self.chek_model('Children', self._build_child(),
                        self.declaration.children)
        # проверяем поля представителя
        conc_delegate, _ = self._build_delegate()
        # тут также надеемся что родитль один и его же мы и редактируем!
        delegate = get_delegate(self.declaration)
        self.chek_model('Delegate', conc_delegate, delegate)

        # проверяем поле ЖУ
        conc_decl_units = self._get_declaration_units_list()
        sys_decl_units = get_declaration_units(self.declaration)
        self.chek_model('DeclarationUnit', conc_decl_units, sys_decl_units)

        self._check_benefits()
        self._check_docs()

        result = self.storage.create_change(
            # Какая-то магия творится с yadicoм
            self.declaration, self.result, self.storage.source)
        if not result:
            msg = u'Diff not found'

        return UpdateApplicationResponse(**{
            'Status': msg
        })


class ResetDeclarationProxy(object):

    __metaclass__ = Injectable

    STATUS_TRANSFER_COMMENT = u"""
        Авто-смена статуса в связи с поступившей заявкой
        на отказ от участия в очереди поступила с ЕПГУ.
        Необходимо уточнить актуальность запроса у законных
        представителей.
        """

    def __init__(self, request, declaration, **kwargs):
        self._request = request
        self.declaration = declaration
        self.result = defaultdict(list)
        self.case_number = kwargs.get("case_number", None)

    def process(self):
        ch_help = DeclarationChangeHelper(Declaration, 'Declaration')
        self.result['Declaration'].extend(
            ch_help.get_status_change(self.declaration))

        ChangeDeclaration(
            declaration=self.declaration,
            data=json.dumps(
                self.result, cls=AdditionalEncoder),
            source=ChangeSource.UPDATE_APPLICATION,
            case_number=self.case_number).save()

        return UpdateApplicationResponse(**{
            'Status': u'Успех'
        })


class IUpdateApplicationProxy(object):

    __metaclass__ = Injectable

    depends_on = ("behavior", "storage", )

    def process(
            self, request, declaration, binary_data,
            request_code, case_number=None):
        behavior = self.behavior(
            request, declaration,
            binary_data=binary_data, request_code=request_code,
            case_number=case_number)

        behavior.storage = self.storage

        return behavior.process()


def new_application_with_duplicate_check(
        ctx, request, binary_data, request_code):
    """Проверяет наличие дубликатов заявки в системе.
    При их отсутствии запускает процесс создания заявки
    :param ctx: Spyne контекст запроса
    :param request: Данные в запросе
    :param binary_data: Бинарники
    :param request_code: код запроса
    :return: Результат работы
    """

    unit_id = request.EduOrganizations.EduOrganization[0].Code
    try:
        unit = Unit.objects.get(id=unit_id)
    except Unit.DoesNotExist:
        raise SpyneException(
            u"Организация с ID={0} не найдена".format(unit_id))

    search_by_name_filter = Q(
        children__surname__iexact=prepare_name(request.DeclaredPerson.LastName),
        children__firstname__iexact=prepare_name(
            request.DeclaredPerson.FirstName),
        children__date_of_birth=request.DeclaredPerson.DateOfBirth,
        mo_id=getattr(unit.get_mo(), "id", None)) & ~Q(
            status__code__in=[DSS.ARCHIVE, DSS.ACCEPTED, DSS.REFUSED])

    try:
        _ = Declaration.objects.get(search_by_name_filter)
    except Declaration.DoesNotExist:
        # Если записи не найдены, то создаем заявку
        result = NewApplicationProxy(
            request, binary_data, request_code).process()
    except Declaration.MultipleObjectsReturned:
        # Если найдено несколько записей, то переходим ко
        # второму этапу проверки
        # Создание фильтра на поиск по документам или по снилс
        search_by_doc_or_snils = Q(
            children__snils=request.DeclaredPerson.Snils) | Q(
                children__dul_series=request.DeclaredPerson.BirthDocSeria,
                children__dul_number=request.DeclaredPerson.BirthDocNumber,
                children__dul_date=request.DeclaredPerson.BirthDocIssueDate)
        # Выборка согласно фильтру
        if Declaration.objects.filter(
                search_by_name_filter,
                search_by_doc_or_snils).exists():
            # Если найдено хоть, что-нибудь, то сообщение об ошибке и выход
            reject_request(
                ctx, u"Заявление по заданным параметрам уже существует")
        else:
            # Если не найдено, то создаем заявку
            result = NewApplicationProxy(
                request, binary_data, request_code).process()
    else:
        reject_request(
            ctx, u"Заявление по заданным параметрам уже существует")

    return result
